import os.path
import shutil
import time

import matplotlib.pyplot as plt
import numpy as np
import pickle
import arithmetics
from graph import act_t, Node, NodePtr

np.set_printoptions(precision=4, suppress=True, linewidth=np.inf)

COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']


class Goofspiel(object):
    def __init__(self):
        # non-iterables
        self.CARD_NUM: int = 4
        self.upcard: list[int] = [1, 1, 1, 1, 1]
        self.weight: dict[str, list[float]] = {"A": [5.00, 1.33, 2.71, 1.80, 2.24], "B": [4.12, 6.28, 3.33, 1.92, 3.60]}
        self.node_layer: list[list[NodePtr]] = []  # nodes in breadth-first hierarchy

        # states
        self.reg: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # reg[p][I][a]
        self.sig: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # sig[p][I][a]
        self.opt: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # opt[p][I][a]
        self.A: dict[NodePtr, np.ndarray] = {}  # self.A[I][i, j]
        self.B: dict[NodePtr, np.ndarray] = {}  # self.B[J][i, j]
        self.cfv: dict[NodePtr, np.ndarray] = {}  # cfv[n][p]
        self.exp: dict[NodePtr, np.ndarray] = {}  # exp[n][p]

        # records
        self.REGs: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # REGs[player][info_set][act, t]
        self.SIGs: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # SIGs[player][info_set][act, t]
        self.OPTs: dict[str, dict[NodePtr, np.ndarray]] = {"A": {}, "B": {}}  # OPTs[player][info_set][act, t]
        self.CFVs: dict[NodePtr, np.ndarray] = {}  # CFVs[node][player, t]
        self.EXPs: dict[NodePtr, np.ndarray] = {}  # EXPs[node][player, t]

    def node_layer_init(self):
        root = Node.new()
        root.o.branch = np.array([0, 0])
        root.o.actA = np.arange(self.CARD_NUM) + 1
        root.o.actB = np.arange(self.CARD_NUM) + 1

        parent = [root]
        child = []
        self.node_layer = []
        while len(parent):
            self.node_layer.append(parent)
            for n in parent:
                n.o.child = np.resize(n.o.child, [len(n.o.actA), len(n.o.actB)])
                for i, a in enumerate(n.o.actA):
                    for j, b in enumerate(n.o.actB):
                        n.o.child[i, j] = Node.new()
                        ch = n.o.child[i, j]
                        ch.o.parent = n
                        ch.o.branch = np.array([a, b])
                        ch.o.actA = n.o.actA[n.o.actA != a]
                        ch.o.actB = n.o.actB[n.o.actB != b]
                        child.append(ch)
            parent = child
            child = []

    def sigma_init(self, T: int = 200):
        for layer in self.node_layer[0:-2]:
            for n in layer:
                self.reg["A"][n] = arithmetics.rand_f(1, len(n.o.actA) - 1)[0] * 5
                self.reg["B"][n] = arithmetics.rand_f(1, len(n.o.actB) - 1)[0] * 5
                self.sig["A"][n] = self.reg["A"][n] / self.reg["A"][n].sum()
                self.sig["B"][n] = self.reg["B"][n] / self.reg["B"][n].sum()

        for layer in self.node_layer[0:-2]:
            for n in layer:
                self.REGs["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.REGs["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
                self.SIGs["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.SIGs["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
        for layer in self.node_layer[0:-2]:
            for n in layer:
                self.REGs["A"][n][:, 0] = self.reg["A"][n]
                self.REGs["B"][n][:, 0] = self.reg["B"][n]
                self.SIGs["A"][n][:, 0] = self.sig["A"][n]
                self.SIGs["B"][n][:, 0] = self.sig["B"][n]

    def cfv_init(self, T: int = 200):
        for layer in self.node_layer:
            for n in layer:
                self.cfv[n] = np.empty([2], dtype=float)
                self.CFVs[n] = np.empty([2, T], dtype=float)
        for layer in self.node_layer[0:-2]:
            for n in layer:
                self.A[n] = np.empty([len(n.o.actA), len(n.o.actB)], dtype=float)
                self.B[n] = np.empty([len(n.o.actA), len(n.o.actB)], dtype=float)
                self.opt["A"][n] = np.empty([len(n.o.actA)], dtype=float)
                self.opt["B"][n] = np.empty([len(n.o.actB)], dtype=float)
                self.exp[n] = np.empty([2], dtype=float)
                self.OPTs["A"][n] = np.empty([len(n.o.actA), T], dtype=float)
                self.OPTs["B"][n] = np.empty([len(n.o.actB), T], dtype=float)
                self.EXPs[n] = np.empty([2, T], dtype=float)

        # leaves - fill the values:
        for n in self.node_layer[-1]:
            h = n.o.h()
            u, v = 0, 0
            for k, acts in enumerate(h):
                a, b = acts
                if a > b:
                    u += self.upcard[k] * self.weight["A"][k]
                    v -= self.upcard[k] * self.weight["B"][k]
                elif a < b:
                    u -= self.upcard[k] * self.weight["A"][k]
                    v += self.upcard[k] * self.weight["B"][k]
            self.cfv[n][:] = np.array([u, v], dtype=float)
            # self.CFVs[n][:, 0] = self.cfv[n].copy()
        for n in self.node_layer[-2]:
            self.cfv[n][:] = self.cfv[n.o.get_child(0, 0)].copy()
            # self.CFVs[n][:, 0] = self.cfv[n].copy()

    def cfv_new(self, t: int):
        '''
        for layer in self.node_layer[-1:-3:-1]:  # leaves and trivial nodes won't change
            for n in layer:
                self.CFVs[n][:, t] = self.CFVs[n][:, 0]
        '''
        for layer in self.node_layer[-3::-1]:
            for n in layer:
                for i, a in enumerate(n.o.actA):
                    for j, b in enumerate(n.o.actB):
                        self.A[n][i, j], self.B[n][i, j] = self.cfv[n.o.get_child(i, j)][:]
                self.opt["A"][n][:] = self.A[n] @ self.sig["B"][n]
                self.opt["B"][n][:] = self.B[n].T @ self.sig["A"][n]
                self.OPTs["A"][n][:, t], self.OPTs["B"][n][:, t] = self.opt["A"][n], self.opt["B"][n]

                self.cfv[n][0] = self.opt["A"][n].T @ self.sig["A"][n]
                self.cfv[n][1] = self.opt["B"][n].T @ self.sig["B"][n]
                self.CFVs[n][0, t], self.CFVs[n][1, t] = self.cfv[n]

                self.exp[n][0] = self.opt["A"][n].max() - self.cfv[n][0]
                self.exp[n][1] = self.opt["B"][n].max() - self.cfv[n][1]
                self.EXPs[n][0, t], self.EXPs[n][1, t] = self.exp[n]
    # New idea: Ditch iterables, rely solely on records! Because this can tackle overwrites! 2022-1121-2330
    # Problem: leaves, trivial nodes? -> Anyway, don't care about compression!
    # Now all iterations are important, so record should only keep part of them, this make record should always be peripheral.

    def sigma_new(self, t: int):
        for layer in self.node_layer[0:-2]:
            for n in layer:
                x = self.sig["A"][n][:]
                y = self.sig["B"][n][:]
                u = self.opt["A"][n][:]
                v = self.opt["B"][n][:]

                r = {"A": u - u.T @ x, "B": v - v.T @ y}
                r["A"][r["A"] < 0] *= 0.05
                r["B"][r["B"] < 0] *= 0.05
                self.reg["A"][n][:] = self.reg["A"][n][:] + r["A"]
                self.reg["B"][n][:] = self.reg["B"][n][:] + r["B"]
                self.reg["A"][n][self.reg["A"][n][:] < 0] = 0
                self.reg["B"][n][self.reg["B"][n][:] < 0] = 0
                self.sig["A"][n][:] = self.reg["A"][n][:] / self.reg["A"][n][:].sum()
                self.sig["B"][n][:] = self.reg["B"][n][:] / self.reg["B"][n][:].sum()

                self.REGs["A"][n][:, t] = self.reg["A"][n][:].copy()
                self.REGs["B"][n][:, t] = self.reg["B"][n][:].copy()
                self.SIGs["A"][n][:, t] = self.sig["A"][n][:].copy()
                self.SIGs["B"][n][:, t] = self.sig["B"][n][:].copy()


def split_to_stackers(A: np.ndarray):
    M, N = A.shape
    y = []
    for i in range(M):
        y.append(A[i, :])
    return y


# noinspection PyUnreachableCode
def init(game: Goofspiel):
    game.node_layer_init()

    T = 501
    game.sigma_init(T)
    game.cfv_init(T)
    game.cfv_new(0)

    with open("Data.db", "wb") as db:
        pickle.dump(Node.N, db)
        pickle.dump(Node.MEM, db)
        pickle.dump(game.node_layer, db)
        Node.N = 16
        Node.MEM = {}
        game.node_layer = []


def run(game: Goofspiel, T=501):
    with open("Data.db", "rb") as db:
        Node.N = pickle.load(db)
        Node.MEM = pickle.load(db)
        game.node_layer = pickle.load(db)

    t_start = time.time()
    for t in range(1, T):
        if t % 50 == 0:
            print(t)
        game.sigma_new(t)
        game.cfv_new(t)
    print("iter:", time.time() - t_start)


def output(game: Goofspiel, T=501):
    for d in ["cfv_A", "cfv_B", "ep", "reg_A", "reg_B", "sig_A", "sig_B"]:
        try:
            shutil.rmtree(d)
        except:
            print("error in rmtree", d)
        finally:
            os.mkdir(d)
    # reg
    t_start = time.time()
    if True:
        for l in game.node_layer[0:-3]:
            for n in l:
                h = str(n.o.h())
                fig = plt.figure(h, [8, 4], 96)
                ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
                colors = [COLORS[a % 10] for a in n.o.actA]
                ax.stackplot(np.arange(T), *split_to_stackers(game.REGs["A"][n]), labels=n.o.actA, colors=colors, alpha=0.5)
                ax.set_title(r'$R^A$' + h)
                ax.grid()
                ax.legend()
                fig.savefig(os.path.join("reg_A", h + ".png"))
                plt.close(fig)

                fig = plt.figure(h, [8, 4], 96)
                ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
                colors = [COLORS[b % 10] for b in n.o.actB]
                ax.stackplot(np.arange(T), *split_to_stackers(game.REGs["B"][n]), labels=n.o.actB, colors=colors, alpha=0.5)
                ax.set_title(r'$R^B$' + h)
                ax.grid()
                ax.legend()
                fig.savefig(os.path.join("reg_B", h + ".png"))
                plt.close(fig)
    print("reg:", time.time() - t_start)

    # sig
    t_start = time.time()
    if True:
        for l in game.node_layer[0:-3]:
            for n in l:
                h = str(n.o.h())
                fig = plt.figure(h, [8, 4], 96)
                ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
                colors = [COLORS[a % 10] for a in n.o.actA]
                ax.stackplot(np.arange(T), *split_to_stackers(game.SIGs["A"][n]), labels=n.o.actA, colors=colors, alpha=0.5)
                ax.set_title(r'$\sigma^A$' + h)
                ax.grid()
                ax.legend()
                fig.savefig(os.path.join("sig_A", h + ".png"))
                plt.close(fig)

                fig = plt.figure(h, [8, 4], 96)
                ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
                colors = [COLORS[b % 10] for b in n.o.actB]
                ax.stackplot(np.arange(T), *split_to_stackers(game.SIGs["B"][n]), labels=n.o.actB, colors=colors, alpha=0.5)
                ax.set_title(r'$\sigma^B$' + h)
                ax.grid()
                ax.legend()
                fig.savefig(os.path.join("sig_B", h + ".png"))
                plt.close(fig)
    print("sig:", time.time() - t_start)

    # cfv
    t_start = time.time()
    if True:
        for l in game.node_layer[0:-3]:
            for n in l:
                h = str(n.o.h())
                fig = plt.figure(h, [8, 4], 96)
                ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
                ax.plot(np.arange(T), game.CFVs[n][0], label="mixed")
                for i, a in enumerate(n.o.actA):
                    ax.plot(np.arange(T), game.OPTs["A"][n][i, :], label=a, alpha=0.3, color="C" + str(a))
                ax.set_title(r'$cfv^A$' + h)
                ax.grid()
                ax.legend()
                fig.savefig(os.path.join("cfv_A", h + ".png"))
                plt.close(fig)

                fig = plt.figure(h, [8, 4], 96)
                ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
                ax.plot(np.arange(T), game.CFVs[n][1], label="mixed")
                for j, b in enumerate(n.o.actB):
                    ax.plot(np.arange(T), game.OPTs["B"][n][j, :], label=b, alpha=0.3, color="C" + str(b))
                ax.set_title(r'$cfv^B$' + h)
                ax.grid()
                ax.legend()
                fig.savefig(os.path.join("cfv_B", h + ".png"))
                plt.close(fig)
    print("cfv:", time.time() - t_start)

    # ep-value
    t_start = time.time()
    if True:
        for l in game.node_layer[0:-3]:
            for n in l:
                h = str(n.o.h())
                fig: plt.Figure = plt.figure(h, [8, 4], 96)
                fig.add_axes([0.1, 0.1, 0.8, 0.8])
                fig.axes[0].loglog(np.arange(T), game.EXPs[n].sum(axis=0), label=r'$ep$')
                fig.axes[0].loglog(np.arange(T), game.EXPs[n][0, :], label=r'$ep_A$', alpha=0.3)
                fig.axes[0].loglog(np.arange(T), game.EXPs[n][1, :], label=r'$ep_B$', alpha=0.3)
                fig.axes[0].set_title(r'$ep$' + h)

                ax: plt.Axes
                for ax in fig.axes:
                    ax.set_ylim(0.01, 10)
                    ax.grid()
                    ax.legend()
                fig.savefig(os.path.join("ep", h + ".png"))
                plt.close(fig)
    print("exp:", time.time() - t_start)


if __name__ == "__main__":
    np.random.seed(133484)
    game = Goofspiel()
    init(game)
    run(game, 501)
    output(game, 501)
